import argparse
from pathlib import Path

import tiler_swift.scripts.quadkey


def main():
    Formatter = argparse.ArgumentDefaultsHelpFormatter

    parser = argparse.ArgumentParser(prog="tiler-swift", description="cuts quadkeys from raster", formatter_class=Formatter)

    arg = parser.add_argument

    arg("raster", type=Path, help="raster to cut quadkeys from")
    arg("-z", "--zoom", type=int, required=True, help="zoom level to cut quadkeys at")
    arg("-o", "--out", type=Path, required=True, help="directory to write quadkeys to")
    arg("-n", "--nodata", help="nodata value to use for filling at edges")

    parser.set_defaults(main=tiler_swift.scripts.quadkey.main)

    args = parser.parse_args()
    args.main(args)


if __name__ == "__main__":
    main()



import math
import logging

import rasterio
import rasterio.crs
import rasterio.warp
import rasterio.coords
import rasterio.windows
import rasterio.transform
import rasterio.vrt

import numpy as np

import mercantile

logger = logging.getLogger("tiler-swift.quadkey")


def quadkeys(dataset, z, nodata=None, tile_filter_fn=None):
    # TODO: we work with one band, we can generalize
    # later; then we have to take care of different
    # nodata values and dtypes on a per band basis
    assert dataset.count == 1, "one band required"

    kwargs = {}

    # WarpedVRT does not work with GCPs by default
    # https://github.com/mapbox/rasterio/issues/2086
    if dataset.transform.is_identity and dataset.gcps:
        kwargs.update({"src_crs": dataset.gcps[1],
                       "src_transform": rasterio.transform.from_gcps(dataset.gcps[0])})

    with rasterio.vrt.WarpedVRT(dataset, crs="EPSG:3857", **kwargs) as vrt:

        # Mercantile requires lng, lat i.e. https://epsg.io/4326
        bounds = rasterio.warp.transform_bounds(vrt.crs, rasterio.crs.CRS({"init": "EPSG:4326"}), *vrt.bounds)

        tiles = mercantile.tiles(*bounds, zooms=[z])

        if tile_filter_fn is not None:
            tiles = (tile for tile in tiles if tile_filter_fn(tile))

        for tile in tiles:
            quadkey = mercantile.quadkey(tile)
            bounds = mercantile.xy_bounds(*tile)

            window = vrt.window(*bounds)

            transform = vrt.window_transform(window.round_lengths(op="ceil"))

            # WarpedVRT does not support boundless reads;
            # work around by manually padding with nodata
            # https://github.com/mapbox/rasterio/issues/2084

            h_orig, w_orig = math.ceil(window.height), math.ceil(window.width)
            col_off_orig, row_off_orig = window.col_off, window.row_off

            window = window.intersection(rasterio.windows.Window(0, 0, vrt.width, vrt.height))

            dcol = int(window.col_off - col_off_orig)
            drow = int(window.row_off - row_off_orig)

            window = window.round_lengths(op="ceil")

            h, w = int(window.height), int(window.width)

            logging.debug(f"Reading quadkey tile data")

            bounded = vrt.read(window=window, out_shape=(dataset.count, h, w))

            # For boundless reads we need a nodata value
            # otherwise we can not fill in at the edges
            # https://github.com/mapbox/rasterio/issues/2087
            dtype = dataset.dtypes[0]
            nodata = nodata if nodata is not None else dataset.nodatavals[0]

            assert nodata is not None, "proper nodata val for dtype is set"

            data = np.full(shape=(dataset.count, h_orig, w_orig),
                           fill_value=nodata, dtype=dtype)

            data[:, drow:drow + h, dcol:dcol + w] = bounded

            profile = vrt.profile
            profile.update({"width": w_orig, "height": h_orig, "nodata": nodata,
                            "dtype": dtype, "transform": transform})

            yield quadkey, data, profile



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tiler_swift', 'tiler_swift.scripts']

package_data = \
{'': ['*']}

install_requires = \
['mercantile>=1.1.6,<2.0.0', 'numpy>=1.19.5,<2.0.0', 'rasterio>=1.1.8,<2.0.0']

entry_points = \
{'console_scripts': ['tiler-swift = tiler_swift.scripts.__main__:main']}

setup_kwargs = {
    'name': 'tiler-swift',
    'version': '0.2.0',
    'description': 'Tiler Swift, the swiftest quadkey tiler in the wild west',
    'long_description': '# tiler-swift\n\nTiler Swift, the swiftest quadkey tiler in the wild west.\n\nFeatures\n- ✔ Get quadkey tiles from raster\n- ✔ Simple and blazingly fast\n\n![](assets/quadkey.jpg)\n\n\n## Table of Contents\n\n1. [Overview](#overview)\n2. [Library](#development)\n3. [Command Line](#command-line)\n4. [Development](#development)\n5. [License](#license)\n\n\n## Overview\n\nThis project provides a Python library and a thin command line wrapper on top of it to cut quadkey raster tiles.\n\nQuadkey tiles are raster files of equal size (width, height) on a specific zoom level; the coordinate reference system is the Mercator Projection.\nThe quadkey identifier is constructed by dividing the world into four, then dividing each cell into four again, and so on, recursively.\n\nSee\n- https://labs.mapbox.com/what-the-tile/\n- https://docs.microsoft.com/en-us/bingmaps/articles/bing-maps-tile-system\n- https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames\n- https://wiki.openstreetmap.org/wiki/Slippy_Map\n\n\n## Library\n\nThe `quadkeys(dataset, z, nodata=0, tile_filter_fn=None)` function provides a generator yielding quadkeys and their associated data.\n\nArgs\n- `dataset`: the rasterio dataset to cut quadkeys from\n- `z`: the zoom level to cut quadkeys at\n- `nodata`: the value to use for filling\n- `tile_filter_fn`: a function `mercantile.Tile -> bool` to filter the quadkey generator with\n\nYields a tuple with\n- `quadkey`: the quadkey name e.g. `1202121012`\n- `data`: the pixel data as a numpy ndarray\n- `profile`: the rasterio profile with metadata e.g. transformation for geo-referencing\n\nExample usage:\n\n```python\nfor quadkey, data, profile in quadkeys(dataset, z=args.zoom):\n    print(quadkey)\n\n    with rasterio.open(args.out, "w", **profile) as out:\n        out.write(data)\n```\n\n\n## Command Line\n\n```\nusage: tiler-swift [-h] -z ZOOM -o OUT [-n NODATA] raster\n\ncuts quadkeys from raster\n\npositional arguments:\n  raster                raster to cut quadkeys from\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -z ZOOM, --zoom ZOOM  zoom level to cut quadkeys at (default: None)\n  -o OUT, --out OUT     directory to write quadkeys to (default: None)\n  -n NODATA, --nodata NODATA\n                        nodata value to use for filling at edges (default: None)\n```\n\nExample usage:\n\n```\nquadkeys input.tif -o output.tif -z 10 -n 0\n```\n\n\n## Development\n\nFor a self-contained and isolated dev environment\n\n    docker-compose build\n\nFor a development shell in the container, run\n\n    docker-compose run --rm --entrypoint bash dev\n    $ quadkey --help\n\nTo run the container manually\n\n    docker run robofarm/quadkey\n\n\n## License\n\nCopyright © 2020 robofarm\n\nDistributed under the MIT License (MIT).\n',
    'author': 'Robofarm',
    'author_email': 'hello@robofarm.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
from pathlib import Path

import rasterio

from tiler_swift.quadkey import quadkeys


def main(args):
    logging.basicConfig(format="%(asctime)s %(message)s", level=logging.INFO)

    logger = logging.getLogger("tiler-swift.quadkey")

    args.out.mkdir(exist_ok=True)

    logger.info(f"Reading raster from {args.raster}")

    with rasterio.open(args.raster) as dataset:
        logger.info(f"Writing quadkeys to {args.out}")

        for quadkey, data, profile in quadkeys(dataset, z=args.zoom, nodata=args.nodata):
            outpath = args.out / Path(quadkey).with_suffix(args.raster.suffix)

            profile.update({"driver": "GTiff", "tiled": False, "compress": "deflate", "predictor": 2})

            with rasterio.open(outpath, "w", **profile) as out:
                out.write(data)
